#include "lib.h"
#include "lib_options.h"

void vdp_copy_from_sys(int index, char* src, int size) {
	char* end = src + size;
	VDP_ADDRESS_REG = index;
	VDP_ADDRESS_REG = (char)(index >> 8) | VDP_WRITE_FLAG;
	while(src < end) VDP_WRITE_DATA_REG = *src++;
}

void vdp_write_byte(int index, char c) {
	VDP_ADDRESS_REG = index;
	VDP_ADDRESS_REG = (char)(index >> 8) | VDP_WRITE_FLAG;
	VDP_WRITE_DATA_REG = c;
}

void vdp_write_reg(int i,char c) {
	VDP_ADDRESS_REG=c;
	VDP_ADDRESS_REG=i|VDP_REG_FLAG;
}

void breakpoint() {
	char* bpt=(char*)0;
	*bpt=0xFF;
}

void interrupts() {
	asm("limi 2");
	asm("limi 0");
}

char random_byte(char c) {
	SEED=SEED*0x6FE5+0x7AB9;
	int i=((unsigned)SEED>>8)+(SEED<<8);
	return (unsigned)i%c;
}

char vdp_read_byte(int index) {
	VDP_ADDRESS_REG = index;
	VDP_ADDRESS_REG = (char)(index >> 8) | VDP_READ_FLAG;
	asm("nop");
	return VDP_READ_DATA_REG;
}

void vdp_read_bytes(int i,char* buf,int n) {
	char* end=buf+n;
	VDP_ADDRESS_REG=i;
	VDP_ADDRESS_REG=(char)(i>>8)|VDP_READ_FLAG;	
	while(buf<end)*buf++=VDP_READ_DATA_REG;
}

int key_scan(char* key) {
	KEY_UNIT=5;
	asm(
		"bl @kscan"
		:::"r11"
	);
	*key=KEY;
	char c=STATUS&0x20;
	return c;
}

void gpl_link(int addr) {
	asm(
		"mov %0,r1\n\t"
		"blwp @gpllnk"
		:
		:"r"(addr)
		:"r1"
	);
}

int strlen(const char* str) {
    const char *s;
    for(s=str;*s;++s);
    return(s-str);
}

void color(int charset,char fg,char bg) {
	char c=bg+(fg<<4);
	vdp_copy_from_sys(charset+0x380,&c,1);
}

void defchar(int i,int* pat) {
	vdp_copy_from_sys(i*8+0x800,(char*)pat,8);
}

void load_standard_characters() {
	FAC=0x900;
	gpl_link(0x18);
	FAC=0xB00;
	gpl_link(0x4A);
	int c[4]= { 0x007C, 0x7C7C, 0x7C7C, 0x7C7C };
	defchar(0x1E,c); // Cursor
	int c2[4]= { 0x0000, 0xFE00, 0x0000, 0x0000 };
	defchar(0x1D,c2); // Underline
	int c3[4]= { 0x0002, 0x060C, 0x5870, 0x2000 };
	defchar(0x1C,c3); // Tick
}

void screen(char color) {
	vdp_write_reg(7,color);
}

void pchar(char c,int y,int x) {
	vdp_copy_from_sys(y*32+x,&c,1);
}

void hchar(char c,int y,int x,int rep) {
	int i;
	for(i=0;i<rep;i++)vdp_copy_from_sys(y*32+x+i,&c,1);
}

void vchar(char c,int y,int x,int rep) {
	int i;
	for(i=0;i<rep;i++)vdp_copy_from_sys((y+i)*32+x,&c,1);
}

void reverse(char s[]) {
	int i,j;
	char c;
	for(i=0,j=strlen(s)-1;i<j;i++,j--) {
	 c = s[i];
	 s[i] = s[j];
	 s[j] = c;
	}
}

void itoa(int n,char s[]) {
	int i,sign;
	if((sign=n)<0)n=-n;
	i=0;
	do { s[i++]=n%10+'0'; } while((n/=10)>0);
	if(sign<0)s[i++]='-';
	s[i]='\0';
	reverse(s);
}

void display_at(int y,int x,char* c) {
	vdp_copy_from_sys(y*32+x,c,strlen(c));
}

void fill_vdp(int addr,int n,char c) {
	VDP_ADDRESS_REG=addr;
	VDP_ADDRESS_REG=(char)(addr>>8)|VDP_WRITE_FLAG;
	int i;
	for(i=0;i<n;i++)VDP_WRITE_DATA_REG=c;
}

void clear() { fill_vdp(0,32*24,' '); }

int accept_keep_buffer(char* c,int l,int y,int x) {
	int cursor=0;
	int rpt=0;
	
	int input_index=0;
	int chr=0;
	int ok=0;
	int right=0;
	int ins=0;
	int x1=x;
	int y1=y;
	
	int i;
	char c2;	
	TIMER=0;
	c[l]='\0';
		
	while(!ok) {
		interrupts();

		if(key_scan(&c2))chr=1;
		else {
			if(KEY!=0xFF) {
				if(!rpt) {
					c2=KEY;
					rpt=1;
					TIMER=0;
					cursor=0;
				}
			}
			else rpt=0;
		}
		
		if(rpt==2 && TIMER>5)chr=1;
		
		if(chr) {
			chr=0;
			TIMER=100;
			cursor=0;
			if(c2>0x0F) {
				if(ins) {
					for(i=l-1;i>input_index;i--)c[i]=c[i-1];
					display_at(y,x,c+input_index);
				}
				c[input_index]=c2;
				pchar(c2,y,x);
				right=1;
			}
			else {
				switch(c2) {
				case 0x0D: // Enter
					ok=1;
					cursor=1;
					TIMER=100;
					break;
				case 0x08: // Left
					ins=0;
					pchar(c[input_index],y,x);
					if(input_index>0) {
						input_index--;
						x--;
						if(x<0) {
							x=31;
							y--;
						}
					}
					break;
				case 0x09: // Right					
					right=1;
					ins=0;
					break;
				case 0x04: // Insert
					ins=1;
					break;
				case 0x03: // Delete
					for(i=input_index;i<=l-1;i++)c[i]=c[i+1];
					c[l-1]=' ';
					display_at(y,x,c+input_index);
					break;
				case 0x07: // Clear
					x=x1; y=y1; input_index=0;
					for(i=0;i<l;i++)c[i]=' ';
					display_at(y,x,c+input_index);
					break;
				}
			}
			
			if(right) {
				right=0;
				pchar(c[input_index],y,x);
				if(input_index<l-1) {
					input_index++;
					x=x+1;
					if(x>31) {
						x=0;
						y=y+1;
					}
				}
			}
		}

		if(cursor) {
			if(TIMER>=20) {
				cursor=0;
				pchar(c[input_index],y,x);
				TIMER=0;
				if(rpt==1)rpt=2;
			}
		}
		else {
			if(TIMER>=15) {
				cursor=1;
				pchar(0x1E,y,x);
				TIMER=0;
			}
		}
	}
	int l2=0;
	for(i=l-1;i>=0;i--) {
		if(c[i]!=' ') {
			l2=i+1;
			break;
		}
	}
	c[l2]=0;
	return l2;
}

int accept(char* c,int l,int y,int x) {
	int i;
	for(i=0;i<l;i++)c[i]='\0';
	accept_keep_buffer(c,l,y,x);
}

int parse_string(char* c,int* v) {
	int i=0;
	int neg=0;
	unsigned int v2=0;
	int ok=0;
	while(*c==' ' || *c=='\t')c++;
	if(*c=='-') {
		neg=1;
		c++;
	}
	while(*c) {
		if(*c>='0' && *c<='9') {
			if(v2>=3276 && (!neg && (*c-'0')>7) || (neg && (*c-'0')>8)) {
				ok=0;
				break;
			}
			else {
				v2=v2*10+(*c-'0');
				c++;
				ok=1;
			}
		}
		else {
			ok=0;
			break;
		}
	}
	if(neg)v2=-v2;
	if(ok)*v=v2;
	return ok;
}

void accept_tone() { gpl_link(0x34); }
void bad_response_tone() { gpl_link(0x36); }

void bye() {
	asm(
		"limi 2\n\t"
		"lwpi >83E0\n\t"
		"blwp @>0"
	);
}

void sound(char* c,int n) {
	vdp_copy_from_sys(0x1800,c,n);
	SOUND_LIST_POINTER=0x1800;
	GPL_R14_LOW=GPL_R14_LOW|1;
	SOUND_PLAYING=1;
}

//*************************************************** SPRITES

#if defined LIB_SPRITES

void position(int i,unsigned char y,unsigned char x) {
	int i2=x+(y<<8);
	vdp_copy_from_sys(0x300+i*4,(char*)&i2,2);
}

void sprite(int i,char character,char color) {
	int i2=color+(character+128<<8);
	vdp_copy_from_sys(0x300+i*4+2,(char*)&i2,2);
}

void defsprite(int i,int* pat) {
	vdp_copy_from_sys(i*8+0x400,(char*)pat,8);
}

void magnify(int i) {
	vdp_write_reg(1,0xE0|(i-1));
}

#endif

//*************************************************** DISPLAY MODES

#if defined LIB_DISPLAY_MODES

void mc_clear() { fill_vdp(0x800,48*62/2,0); }

void mc_point(char c,int y,int x) {
	int a=x/2*16+x%2+y/8*512+y%8*2;
	int odd=a&1;
	a=a/2+0x800;
	char c2=vdp_read_byte(a);
	if(odd)vdp_write_byte(a,(c2&0xF0)+c);
	else vdp_write_byte(a,(c2&0x0F)+(c<<4));
}

void init_vdp_registers() {
	vdp_write_reg(0,0x00);
	vdp_write_reg(2,0xF0);
	vdp_write_reg(3,0x0E);
	vdp_write_reg(4,0xF9);
}

void multicolor_mode() {
	init_vdp_registers();
	vdp_write_reg(1,0xE8);
	mc_clear();
	
	char c[32];
	int i;
	int v=0;
	int i2=0;
	int i3;
	int i4=0;
	
	while(i4<6) {
		i3=0;
		for(i=0;i<32;i++) {
			c[i]=v;
			v++;
		}
		while(i3<4) {
			vdp_copy_from_sys(i2,c,32);
			i2+=32;
			i3++;
		}
		i4++;
	}	

}

void graphics_mode() {
	init_vdp_registers();
	vdp_write_reg(1,0xE0);
	clear();
	load_standard_characters();
	fill_vdp(0x300,0xC0,0);
}

void bmp_clear() { fill_vdp(0x2000,0x1800,0); }

void bitmap_mode() {
	vdp_write_reg(0,0x02);
	vdp_write_reg(2,0x0E);
	vdp_write_reg(3,0xFF);
	vdp_write_reg(4,0x03);

	//Screen table	
	int i=0x3800;
	VDP_ADDRESS_REG=i;
	VDP_ADDRESS_REG=(char)(i>>8)|VDP_WRITE_FLAG;
	int i2;
	for(i2=1;i2<=3;i2++)
		for(i=0;i<0x100;i++)VDP_WRITE_DATA_REG=i;
		
	//Pattern table
	fill_vdp(0,0x1800,0xF0);
	
	bmp_clear();
}

void bmp_nibble(char c,int y,int x) {
	int a=x/2*16+x%2+y/8*512+y%8*2;
	int odd=a&1;
	a=a/2+0x2000;
	char c2=vdp_read_byte(a);
	if(odd)vdp_write_byte(a,(c2&0xF0)+c);
	else vdp_write_byte(a,(c2&0x0F)+(c<<4));
}

void text_clear() { fill_vdp(0,40*24,' '); }

void text_mode() {
	init_vdp_registers();
	vdp_write_reg(1,0xF0);
	vdp_write_reg(7,0x17);
	text_clear();
	load_standard_characters();
}

void text_pchar(char c,int y,int x) {
	vdp_copy_from_sys(y*40+x,&c,1);
}

#endif

//*************************************************** FILES

#if defined LIB_FILES

int dsr_link(int op) {
	volatile int stat;
	asm(
		"mov %1,r1\n\t"
		"blwp @dsrlnk\n\t"
		"stst %0"
		:"=r"(stat)
		:"r"(op)
		:"r0","r1"
	);
	return stat;
}

int file_operation(pab* p,int op) {
	p->opcode=op;
	int EOF=0;
	p->flags=p->flags&0x1F;
	vdp_copy_from_sys(0x3B02,(char*)p,sizeof(pab));
	PAB_NAMELENGTH_ADDR=0x3B0B;
	int stat=dsr_link(8);
	vdp_read_bytes(0x3B02,(char*)p,sizeof(pab));
	int err=p->flags>>5;
	if(err==5) {
		EOF=1;
		err=0;
	}
	else {
		if(err)err++;
		else if(stat&0x2000)err=1;
	}
	
	if(err && p->opcode>1)close(p);
	return err;	
}

char *strcpy(char *dest,const char *src) {
   char *save=dest;
   while(*dest++=*src++);
   return save;
}

int open(pab* p,const char* s,char mode) {
	p->flags=0x10|mode;
	p->buffer_address=0x3B80;
	p->record_length=80;
	strcpy(p->name,s);
	p->name_length=strlen(p->name);
	return file_operation(p,0);
}

int open_with_options(pab* p,const char* s,char mode,int vdp,char data_type,char record_type,char record_length) {
	p->flags=mode|data_type|record_type;
	p->buffer_address=vdp;
	p->record_length=record_length;
	strcpy(p->name,s);
	p->name_length=strlen(p->name);
	return file_operation(p,0);
}

int write(pab* p,char* buf,char n) {
	vdp_copy_from_sys(0x3B80,buf,n);
	p->character_count=n;
	return file_operation(p,3);
}

int read(pab* p,char* buf) {
	int ret=file_operation(p,2);
	vdp_read_bytes(0x3B80,buf,p->character_count);
	return ret;
}

void close(pab* p) { file_operation(p,1); }

int save_from_vdp(pab* p,const char* s,int vdp,int n) {
	p->record_number=n;
	p->buffer_address=vdp;
	strcpy(p->name,s);
	p->name_length=strlen(p->name);
	return file_operation(p,6);
}

int save(pab* p,const char* s,char* buf,int n) {
	vdp_copy_from_sys(0x1000,buf,n);
	return save_from_vdp(p,s,0x1000,n);
}

int load_to_vdp(pab* p,const char* s,int vdp,int n) {
	p->record_number=n;
	p->buffer_address=vdp;
	strcpy(p->name,s);
	p->name_length=strlen(p->name);
	return file_operation(p,5);
}

int load(pab* p,const char* s,char* buf,int n) {
	int ret=load_to_vdp(p,s,0x1000,n);
	vdp_read_bytes(0x1000,buf,n);
	return ret;
}

#endif

//*************************************************** TEXT DISPLAY

#if defined LIB_TEXT_DISPLAY

void format_string(char* c,char* f,...) {
	va_list l;
	va_start(l,f);
	int i;
	int i2=0;
	for(i=0;i<strlen(f);i++) {
		if(f[i]!='%')c[i2++]=f[i];
		else {
			char s7[7];
			itoa(va_arg(l,int),s7);
			int i3;
			for(i3=0;i3<strlen(s7);i3++)c[i2++]=s7[i3];
		}
	}
	c[i2]='\0';
	va_end(l);
}

void clear_rectangle(int y,int x,int h, int w) {
	int i;
	for(i=y;i<y+h;i++)hchar(' ',i,x,w);
}

int display_at_multiline (int y,int x,char* s,int w) {
	int i1=0;
	int n;
	int x2=x;
	int y2=y;
	int l=strlen(s);
	bool b2=false;
	do {
		bool b;
		n=0;
		do {
			if(*(s+i1+n)==' ') {
				b=false;
				break;
			}
			if(*(s+i1+n)=='\n') {
				b=true;
				break;
			}
			if(*(s+i1+n)=='\0') {
				b=false;
				break;
			}
			n++;
		} while(1);
		
		if(x2+n>w+x || b2) {
			hchar(' ',y2,x2,w-x2);
			x2=x;
			y2++;
		}
		b2=b;
		vdp_copy_from_sys(y2*32+x2,s+i1,n);
		x2+=n;
		if(x2<w+x) {
			pchar(' ',y2,x2);
			x2++;
		}
		i1+=n+1;		
	} while(i1<l);
	if(x2<w+x)hchar(' ',y2,x2,w-x2);
	return y2-y+1;
}

#endif

//*************************************************** BASIC DIALOGS

#if defined LIB_BASIC_DIALOGS

void edit_action(void* v) {
	accept_tone();
	Edit* e=(Edit*)((Field*)v)->object;	
	bool ok=false;
	while(!ok) {
		interrupts();		
		char s7[7];
		if(e->numeric)itoa(e->value,s7);
		Field* f=&e->field;
		point* p=&f->position;
		hchar(' ',p->y,f->value_x,e->length);
		char* s;
		if(e->numeric)s=s7;
		else s=e->buffer;
		
		display_at(p->y,f->value_x,s);
		int i;
		for(i=strlen(s);i<e->length;i++)s[i]=' ';
		
		accept_keep_buffer(s,e->length,p->y,f->value_x);
		if(e->numeric) {
			if(parse_string(s7,&e->value)) {
				if(e->value < e->min || e->value > e->max) {
					char s40[40];
					format_string(s40,"Minimum: %, Maximum: %",e->min,e->max);
					hchar(' ',23,0,32);
					display_at(23,2,s40);
					bad_response_tone();
				}
				else ok=true;
			}
			else bad_response_tone();
		}
		else ok=true;
	}
}

void show_field_text(Field* f) {
	display_at(f->position.y,f->position.x,f->text);
	hchar(' ',f->position.y+1,f->position.x,strlen(f->text));
	pchar(0x1D,f->position.y+1,f->position.x+f->key_index);
}

void edit_show(void* v) {
	Field* f=(Field*)v;
	Edit* e=(Edit*)f->object;
	show_field_text(f);
	char* s2;
	char s[7];
	if(e->numeric) {
		itoa(e->value,s);
		s2=s;
	}
	else s2=e->buffer;
	
	display_at(f->position.y,f->value_x,s2);
}

void create_field(void* v,Field* f,void(*s)(void*),void(*a)(void*)) {
	f->object=v;
	f->show=s;
	f->action=a;
	f->value_x=strlen(f->text)+f->position.x;
}

void create_edit(Edit* e) {
	create_field(e,&e->field,edit_show,edit_action);
}

void button_show(void* v) {
	Field* f=(Field*)v;
	show_field_text(f);	
}

void create_button(Button* b) {
	b->field.show=button_show;
}

void show_context_menu_selection(ContextMenu* m) {
	int i;
	int i2=item_index(m);
	for(i=0;i<m->items_count;i++) {
		if(i2==i)
			pchar(0x1C,m->position.y+i,m->position.x);
		else pchar(' ',m->position.y+i,m->position.x);
	}
}

int item_index(ContextMenu* m) {
	int i;
	for(i=0;i<m->items_count;i++)if(m->items[i].value==m->value)return i;
	return -1;
}

void show_context_menu(ContextMenu* m) {
	int i,w1;
	int w=0;
	for(i=0;i<m->items_count;i++) {
		w1=strlen(m->items[i].text);
		if(w1>w)w=w1;
	}
	
	for(i=0;i<m->items_count;i++) {
		hchar(' ',m->position.y+i,m->position.x+1,w);
		display_at(m->position.y+i,m->position.x+1,m->items[i].text);
	}
	
	show_context_menu_selection(m);
	
	bool ok=false;
	int index=item_index(m);
	while(!ok) {
		interrupts();
		char key;
		if(key_scan(&key)) {
			switch(key) {
			case DOWN:
			case 'X':
				index++;
				if(index>=m->items_count)index=0;
				m->value=m->items[index].value;
				show_context_menu_selection(m);
				break;
			case UP:
			case 'E':
				index--;
				if(index<0)index=m->items_count-1;
				m->value=m->items[index].value;
				show_context_menu_selection(m);
				break;
			case 0x0D:
				ok=true;
			}
		}		
	}
	clear_rectangle(m->position.y,m->position.x,m->items_count,w+1);
}

#endif

//*************************************************** EXTENDED DIALOGS

#if defined LIB_EXTENDED_DIALOGS

void combo_box_show(void* v) {
	Field* f=(Field*)v;
	ComboBox* c=(ComboBox*)f->object;
	show_field_text(f);
	display_at(f->position.y,f->value_x,c->menu.items[item_index(&c->menu)].text);
}

void combo_box_action(void* v) {
	Field* f=(Field*)v;
	ComboBox* c=(ComboBox*)f->object;
	show_context_menu(&c->menu);
	combo_box_show(&c->field);
}

void create_combo_box(ComboBox* c) {
	create_field(c,&c->field,combo_box_show,combo_box_action);
	c->menu.position.x=c->field.value_x;
	c->menu.position.y=c->field.position.y;
}

#endif