#include "stdarg.h"

#define VDP_READ_DATA_REG	(*(volatile char*)0x8800)
#define VDP_WRITE_DATA_REG	(*(volatile char*)0x8C00)
#define VDP_ADDRESS_REG		(*(volatile char*)0x8C02)
#define VDP_READ_FLAG		0x00
#define VDP_WRITE_FLAG		0x40
#define VDP_REG_FLAG		0x80
#define SEED			(*(int*)0x83C0)
#define STATUS			(*(char*)0x837C)
#define KEY_UNIT		(*(char*)0x8374)
#define KEY			(*(volatile unsigned char*)0x8375)
#define JOYST_X			(*(char*)0x8376)
#define JOYST_Y			(*(char*)0x8377)
#define FAC			(*(volatile int*)0x834A)
#define PAB_NAMELENGTH_ADDR	(*(volatile int*)0x8356)
#define INPUT			0x4
#define OUTPUT			0x2
#define UPDATE			0x0
#define DISPLAY			0x0
#define INTERNAL		0x8
#define FIXED_LENGTH		0x0
#define VARIABLE_LENGTH		0x10
#define SEQUENTIAL		0x0
#define RELATIVE		0x1
#define SOUND_PLAYING		(*(char*)0x83CE)
#define SOUND_LIST_POINTER	(*(int*)0x83CC)
#define GPL_R14_LOW		(*(char*)0x83FD)
#define TIMER			(*(volatile char*)0x8379)
#define USER_INTERRUPT		(*(int*)0x83C4)

#define TRANSPARENT		0
#define BLACK			1
#define MEDIUM_GREEN		2
#define LIGHT_GREEN		3
#define	DARK_BLUE		4
#define LIGHT_BLUE		5
#define DARK_RED		6
#define CYAN			7
#define MEDIUM_RED		8
#define LIGHT_RED		9
#define DARK_YELLOW		10
#define LIGHT_YELLOW		11
#define DARK_GREEN		12
#define MAGENTA			13
#define GRAY			14
#define WHITE			15

#define LEFT			8
#define RIGHT			9
#define DOWN			10
#define UP			11

#define true			0xFFFF
#define false			0
typedef int bool;

typedef struct {
	int x;
	int y;
} point;

void vdp_copy_from_sys(int index, char* src, int size);
void vdp_write_byte(int index, char c);
void vdp_write_reg(int i,char c);
void breakpoint();
void interrupts();
char random_byte(char c);
char vdp_read_byte(int index);
void vdp_read_bytes(int i,char* buf,int n);
int key_scan(char* key);
void gpl_link(int addr);
int strlen(const char* str);
void color(int charset,char fg,char bg);
void defchar(int i,int* pat);
void load_standard_characters();
void screen(char color);
void pchar(char c,int y,int x);
void hchar(char c,int y,int x,int rep);
void vchar(char c,int y,int x,int rep);
void reverse(char s[]);
void itoa(int n,char s[]);
void display_at(int y,int x,char* c);
void fill_vdp(int addr,int n,char c);
void clear();
int accept(char* c,int l,int y,int x);
int parse_string(char* c,int* v);
void accept_tone();
void bad_response_tone();
void bye();
void sound(char* c,int n);

//*************************************************** SPRITES

void position(int i,unsigned char y,unsigned char x);
void sprite(int i,char character,char color);
void defsprite(int i,int* pat);
void magnify(int i);

//*************************************************** DISPLAY MODES

void mc_clear();
void mc_point(char c,int y,int x);
void init_vdp_registers();
void multicolor_mode();
void graphics_mode();
void bmp_clear();
void bitmap_mode();
void bmp_nibble(char c,int y,int x);
void text_clear();
void text_mode();
void text_pchar(char c,int y,int x);

//*************************************************** FILES

typedef struct {
	char opcode;
	unsigned volatile char flags;
	int buffer_address;
	char record_length;
	char character_count;
	int record_number;
	char screen_offset;
	char name_length;
	char name[50];
} pab;

int dsr_link(int op);
int file_operation(pab* p,int op);
char *strcpy(char *dest,const char *src);
int open(pab* p,const char* s,char mode);
int open_with_options(pab* p,const char* s,char mode,int vdp,char data_type,char record_type,char record_length);
int write(pab* p,char* buf,char n);
int read(pab* p,char* buf);
void close(pab* p);
int save_from_vdp(pab* p,const char* s,int vdp,int n);
int save(pab* p,const char* s,char* buf,int n);
int load_to_vdp(pab* p,const char* s,int vdp,int n);
int load(pab* p,const char* s,char* buf,int n);

//*************************************************** TEXT DISPLAY

void format_string(char* c,char* f,...);
void clear_rectangle(int y,int x,int h, int w);
int display_at_multiline (int y,int x,char* s,int w);

//*************************************************** DIALOG LIBRARY

typedef struct {
	void (*action)(void*);
	char key;
	void* object;
	point position;
	char* text;
	void(*show)(void*);
	int key_index,value_x;
} Field;

typedef struct {
	Field field;
} Button;

typedef struct {
	Field field;
	int value,min,max,length;
	bool numeric;
	char* buffer;
} Edit;

typedef struct {
	char* text;
	int value;
} MenuItem;

typedef struct {
	MenuItem* items;
	int items_count;
	point position;
	int value;
} ContextMenu;

typedef struct {
	Field field;
	ContextMenu menu;
} ComboBox;

void edit_action(void* v);
void show_field_text(Field* f);
void edit_show(void* v);
void create_field(void* v,Field* f,void(*s)(void*),void(*a)(void*));
void create_edit(Edit* e);
void button_show(void* v);
void create_button(Button* b);
void show_context_menu_selection(ContextMenu* m);
int item_index(ContextMenu* m);
void show_context_menu(ContextMenu* m);
void combo_box_show(void* v);
void combo_box_action(void* v);
void create_combo_box(ComboBox* c);
