#include "lib.h"

#define MULTICOLOR_MODE	0
#define BITMAP_MODE	1
#define	GRAPHICS_MODE	2
#define	TEXT_MODE	3

char* pixels=(char*)0xD000;
char* memory=(char*)0xE800;
int colors_number=8;
int mode=MULTICOLOR_MODE;
int width,height;
void(*set_point)(char c,int y,int x);

void set_pixel(char* c,int i,char col) {
	int i2=i>>1;
	if(i&1)c[i2]=(c[i2]&0x0F)+(col<<4);
	else c[i2]=(c[i2]&0xF0)+col;
}
char get_pixel(char* c,int i) {
	int i2=i>>1;
	if(i&1)return (unsigned char)c[i2]>>4;
	else return c[i2]&0x0F;
}

void random_pixels() {
	int i;
	for(i=0;i<width*height;i++)set_pixel(pixels,i,random_byte(colors_number)+1);
}

int show_pixels() {
	int x,y;
	char c;
	for(y=0;y<height;y++) {
		for(x=0;x<width;x++) {
			set_point(get_pixel(pixels,y*width+x),y,x);
		}
		if(!(y%5)) {
			interrupts();		
			if(key_scan(&c))return 1;
		}
	}
	return 0;
}

	
void test_pixel(int x,int y) {
	char c1,c2,c3;
	int x2,y2;
	
	c1=get_pixel(memory,y*width+x);
	c3=c1+1;
	if(c3>colors_number)c3=1;
	
	x2=x+1;
	if(x2>width-1)x2=0;
	y2=y;
	c2=get_pixel(memory,y2*width+x2);
	if(c2==c3)set_pixel(pixels,y2*width+x2,c1);
	
	x2=x;
	y2=y+1;
	if(y2>height-1)y2=0;
	c2=get_pixel(memory,y2*width+x2);
	if(c2==c3)set_pixel(pixels,y2*width+x2,c1);
	
	x2=x-1;
	if(x2<0)x2=width-1;
	y2=y;
	c2=get_pixel(memory,y2*width+x2);
	if(c2==c3)set_pixel(pixels,y2*width+x2,c1);
	
	x2=x;
	y2=y-1;	
	if(y2<0)y2=height-1;
	c2=get_pixel(memory,y2*width+x2);
	if(c2==c3)set_pixel(pixels,y2*width+x2,c1);
}

int demons() {
	int x,y;
	char c;
	for(y=0;y<height;y++) {
		for(x=0;x<width;x++) {
			test_pixel(x,y);
		}
		if(!(y%5)) {
			interrupts();		
			if(key_scan(&c))return 1;
		}
	}
	return 0;
}

bool close_dialog;
bool dialog_ok;

void ok_action(void* v) {
	close_dialog=true;
	dialog_ok=true;
}

void cancel_action(void* v) {
	close_dialog=true;
	dialog_ok=false;
}

int dialog() {
	graphics_mode();
	int i;
	for(i=3;i<=15;i++)color(i,BLACK,CYAN);

	Button b1;
	b1.field.text="[   OK   ]";
	b1.field.key='O';
	b1.field.key_index=4;
	b1.field.position.x=2;
	b1.field.position.y=21;
	b1.field.action=ok_action;
	create_button(&b1);

	Button b2;
	b2.field.text="[ Cancel ]";
	b2.field.key='C';
	b2.field.key_index=2;
	b2.field.position.x=20;
	b2.field.position.y=21;
	b2.field.action=cancel_action;
	create_button(&b2);

	Edit e;
	e.field.text="Number of colors: ";
	e.field.key='N';
	e.field.key_index=0;
	e.field.position.x=2;
	e.field.position.y=5;
	e.value=colors_number;
	e.min=2;
	e.max=15;
	e.length=3;
	create_edit(&e);
	
	ComboBox c;
	c.field.text="Display: ";
	c.field.key='D';
	c.field.key_index=0;
	c.field.position.x=2;
	c.field.position.y=7;	
	c.menu.value=mode;
	MenuItem menus[4]={ {"Multicolor Mode",0} , {"Bitmap Mode",1} , {"Graphics Mode",2}, {"Text Mode",3} };
	MenuItem items[4];
	c.menu.items_count=4;
	c.menu.items=menus;
	create_combo_box(&c);

	int fields_number=4;
	Field* fields[4]= {&b1.field,&b2.field,&e.field,&c.field};

	char* s1="----- DEMONS -----";
	display_at(1,7,s1);
	for(i=0;i<fields_number;i++)fields[i]->show(fields[i]);
	
	char key;
	close_dialog=false;
	while(!close_dialog) {
		interrupts();
		if(key_scan(&key)) {
			for(i=0;i<fields_number;i++) {
				Field* f=fields[i];
				if(key==f->key)f->action(f);
			}
		}
	}

	int result=0;
	if(dialog_ok) {
		result=1;
		colors_number=e.value;
		mode=c.menu.value;
	}
	return result;
}

void set_graphics_point(char c,int y,int x) {
	pchar(c*8,y,x);
}

int text_patterns[10][4]= {
	{ 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF },
	{ 0xDDFF, 0x77FF, 0xDDFF, 0x77FF },
	{ 0xEEDD, 0xBB77, 0xEEDD, 0xBB77 },
	{ 0x55FF, 0x55FF, 0x55FF, 0x55FF },
	{ 0xDD77, 0xDD77, 0xDD77, 0xDD77 },
	{ 0xAA55, 0xAA55, 0xAA55, 0xAA55 },
	{ 0xAA00, 0xAA00, 0xAA00, 0xAA00 },
	{ 0x8800, 0x2200, 0x8800, 0x2200 },
	{ 0x8800, 0x0000, 0x8800, 0x0000 },
	{ 0x0000, 0x0000, 0x0000, 0x0000 }	
};

void set_text_patterns() {
	int i;
	for(i=0;i<10;i++)
		defchar(i,text_patterns[i]);
	for(i=0;i<5;i++)
		defchar(i+10,text_patterns[i]);
}

void main() {	
	screen(CYAN);
	
	int result=1;
	while(1) {
		switch(mode) {
		case MULTICOLOR_MODE:
			multicolor_mode();
			width=64;
			height=48;
			set_point=mc_point;
			break;
		case BITMAP_MODE:
			bitmap_mode();
			width=64;
			height=48*4;
			set_point=bmp_nibble;
			break;
		case GRAPHICS_MODE:
			width=32;
			height=24;
			set_point=set_graphics_point;
			int i;
			int c[4]={0,0,0,0};
			for(i=0;i<15;i++) {
				color(i,BLACK,i);
				defchar(i*8,c);
			}
			break;
		case TEXT_MODE:
			text_mode();
			set_text_patterns();
			width=40;
			height=24;
			set_point=text_pchar;
			break;
		}

		if(result==1)random_pixels();
		int key=show_pixels();
		
		while(!key) {
			int i;
			for(i=0;i<width*height/2;i++)memory[i]=pixels[i];
			key=demons();
			if(!key)key=show_pixels();
		}
		
		result=dialog();
	}	
}